#include "fw/reactor.h"
#include "test/servo.h"

#include <schpin_lib_exp/config.h>
#include <schpin_msgs/BlobCmd.h>

using namespace schpin;

struct sim_config
{
        std::map< std::string, float > default_angles;

        static auto ros_param_get()
        {
                return std::make_tuple(
                    ros_param< std::map< std::string, float > >( "default_angles",
                                                                 "Default angles at the start of "
                                                                 "the simulation" ) );
        }
};

class reactor_simulation
{
        sim_config                            c_;
        ros::NodeHandle                       handle_{ "~" };
        blob_reactor< 20 >                    reactor_;
        ros::Publisher                        pub_;
        ros::Subscriber                       sub_;
        std::map< lewan_servo_id, sim_servo > servos_;
        ctl_id                                id_            = 0;
        s_ctl_id                              servo_counter_ = 0;

      public:
        reactor_simulation( const sim_config& c )
          : c_( c )
          , pub_( handle_.advertise< schpin_msgs::BlobCmd >( "/blob/feedback", 200 ) )
          , sub_( handle_.subscribe( "/blob/control",
                                     200,
                                     &reactor_simulation::control_topic,
                                     this,
                                     ros::TransportHints().tcpNoDelay() ) )
        {
                for ( auto [id, angle] : c.default_angles ) {
                        auto opt_id =
                            lewan_servo_id::make( static_cast< uint8_t >( std::stoi( id ) ) );
                        if ( opt_id ) {
                                SCHPIN_INFO_LOG( "sim", "Servo: " << id << " default: " << angle );
                                servos_[*opt_id] = sim_servo{ *opt_id, angle };
                        } else {
                                SCHPIN_ERROR_LOG( "sim",
                                                  "ID " << id
                                                        << " in config is not valid for "
                                                           "servo_id " );
                        }
                }
        }

        auto make_matcher()
        {
                return match(
                    [this]( control_tag, fw_node_group_message group_msg ) {
                            auto m = fw_node_protocol::serialize( id_, group_msg, checksum{} );

                            schpin_msgs::BlobCmd msg;
                            msg.data = map_f_to_v( m, []( uint8_t val ) {
                                    return val;
                            } );
                            pub_.publish( msg );
                            id_++;
                    },
                    [&]( servo_tag, s_ctl_id msg_id, reply, const master_lewan_message& m ) {
                            master_lewan_protocol::extract( m )
                                .bind_left(
                                    em::uncurry |
                                    [&]( lewan_servo_id id, master_lewan_submessage submsg ) {
                                            return master_lewan_group::extract( submsg )
                                                .convert_left( [&]( auto var ) {
                                                        return std::make_tuple( id, var );
                                                } );
                                    } )
                                .match(
                                    em::uncurry |
                                        [&]( lewan_servo_id id, master_lewan_variant v ) {
                                                pass_to_sim( msg_id, id, v );
                                        },
                                    [&]( auto e ) {
                                            SCHPIN_ERROR_LOG( "sim", e );
                                    } );
                    } );
        }

        void pass_to_sim( s_ctl_id s_id, lewan_servo_id id, master_lewan_variant v )
        {
                auto opt_msg = servos_[id].on_cmd( get_system_time(), id, v );
                if ( opt_msg ) {
                        reactor_.on_servo_receive(
                            get_system_time(), s_id, *opt_msg, make_matcher() );
                }
        }

        void control_topic( const schpin_msgs::BlobCmd::ConstPtr& msg )
        {
                auto opt_msg = node_fw_message::make( msg->data );
                if ( !opt_msg ) {
                        SCHPIN_ERROR_LOG( "node",
                                          "Data that arrived from control topic are too long" );
                        return;
                }
                parse_node_fw( *opt_msg )
                    .match(
                        em::uncurry |
                            [&]( ctl_id id, node_fw_group_variant subvar, checksum ) {
                                    reactor_.on_control_receive( id, subvar, make_matcher() );
                            },
                        [&]( auto e ) {
                                SCHPIN_ERROR_LOG( "sim", e );
                        } );
        }

        void tick()
        {
                reactor_.tick( get_system_time(), make_matcher() );
        }
};

int main( int argc, char* argv[] )
{
        ros::init( argc, argv, "simulation_node" );

        get_config< sim_config >( "~" ).match(
            [&]( const sim_config& config ) {
                    reactor_simulation sim{ config };

                    ros::Rate r( 600 );
                    while ( ros::ok() ) {
                            sim.tick();
                            ros::spinOnce();
                            r.sleep();
                    }
            },
            [&]( const config_error& e ) {
                    std::cerr << e;
            } );
}
