#include <schpin_lib_exp/bounded.h>
#include <schpin_lib_exp/protocol/item.h>
#include <schpin_lib_exp/protocol/sizeless_message.h>

#pragma once

namespace schpin
{

struct dynmx_error
{
        bool overload;
        bool electrical_shock;
        bool motor_encoder;
        bool overheating;
        bool input_voltage;
};

struct dynmx_status
{
        enum profile_enum
        {
                TRAPEZOIADL_PROFILE = 0b11,
                TRIANGULAR_PROFILE  = 0b10,
                RECTANGULAR_PROFILE = 0b01,
                NO_PROFILE          = 0b00
        };

        profile_enum profile;
        bool         following;
        bool         ongoing;
        bool         arrived;
};

using dynmx_header                           = bounded< uint32_t, 0xFFFFFD00, 0xFFFFFD00 >;
static constexpr dynmx_header dynmx_header_v = dynmx_header::get< 0xFFFFFD00 >();
using dynmx_servo_id                         = bounded< uint8_t, 0, 252 >;
using dynmx_packet_size                      = uint16_t;
using dynmx_crc                              = uint16_t;

using dynmx_response_error = bounded< uint8_t, 0x0, 0x7 >;  // maybe enum here
static constexpr dynmx_response_error DYNMX_NO_ERROR         = dynmx_response_error::get< 0x00 >();
static constexpr dynmx_response_error DYNMX_RESULT_ERROR     = dynmx_response_error::get< 0x01 >();
static constexpr dynmx_response_error DYNMX_INSTRU_ERROR     = dynmx_response_error::get< 0x02 >();
static constexpr dynmx_response_error DYNMX_CRC_ERROR        = dynmx_response_error::get< 0x03 >();
static constexpr dynmx_response_error DYNMX_DATA_RANGE_ERROR = dynmx_response_error::get< 0x04 >();
static constexpr dynmx_response_error DYNMX_DATA_LEN_ERROR   = dynmx_response_error::get< 0x05 >();
static constexpr dynmx_response_error DYNMX_DATA_LIM_ERROR   = dynmx_response_error::get< 0x06 >();
static constexpr dynmx_response_error DYNMX_ACCESS_ERROR     = dynmx_response_error::get< 0x07 >();

using dynmx_model_number  = uint16_t;
using dynmx_model_info    = uint16_t;
using dynmx_fw_version    = uint8_t;
using dynmx_baud_rate     = bounded< uint8_t, 0, 7 >;
using dynmx_delay_time    = bounded< uint8_t, 0, 254 >;
using dynmx_drive_mode    = bounded< uint8_t, 0, 5 >;
using dynmx_oper_mode     = bounded< uint8_t, 0, 16 >;
using dynmx_proto_type    = bounded< uint8_t, 1, 2 >;
using dynmx_home_offset   = bounded< int32_t, -1'044'479, 1'044'479 >;
using dynmx_vel_limit     = bounded< uint32_t, 0, 1'023 >;
using dynmx_voltage       = bounded< uint16_t, 60, 140 >;
using dynmx_temp          = bounded< uint8_t, 0, 100 >;
using dynmx_pwm_limit     = bounded< uint16_t, 0, 885 >;
using dynmx_pos           = bounded< uint32_t, 0, 4'095 >;
using dynmx_bool          = bounded< uint8_t, 0, 1 >;
using dynmx_status_level  = bounded< uint8_t, 0, 2 >;
using dynmx_pid_gain      = bounded< uint16_t, 0, 16'383 >;
using dynmx_watchdog      = bounded< int16_t, -1, 127 >;
using dynmx_pwm           = bounded< int16_t, -885, 885 >;
using dynmx_vel           = bounded< int32_t, -1'023, 1'023 >;
using dynmx_profile_accel = bounded< uint32_t, 0, 32'737 >;
using dynmx_profile_vel   = bounded< uint32_t, 0, 32'767 >;
using dynmx_time          = bounded< uint16_t, 0, 32'767 >;
using dynmx_load          = bounded< int16_t, -1'000, 1'000 >;
using dynmx_addr          = uint16_t;
using dynmx_len           = uint16_t;
using dynmx_buffer        = prot_sizeless_message< 16 >;
using dynmx_reset_flags   = bounded< uint8_t, 0x01, 0x02 >;  // missing 0xFF option, maybe new type?

template <>
struct protocol_item< dynmx_error > : protocol_item_base< dynmx_error >
{
        using value_type = dynmx_error;

        static constexpr std::size_t size = 1;

        static std::size_t serialize_at( static_view< size, uint8_t* > buffer,
                                         const dynmx_error&            e )
        {
                std::bitset< 8 > bset;
                bset[5] = e.overload;
                bset[4] = e.electrical_shock;
                bset[3] = e.motor_encoder;
                bset[2] = e.overheating;
                bset[0] = e.input_voltage;
                return protocol_item< std::bitset< 8 > >::serialize_at( buffer, bset );
        }
        template < typename Buffer, typename Visitor >
        static constexpr std::size_t deserialize( Buffer&& buffer, Visitor&& vis )
        {
                return protocol_item< std::bitset< 8 > >::deserialize(
                    buffer,
                    match(
                        [&]( value_tag, std::bitset< 8 > bset ) {
                                dynmx_error de;
                                de.overload         = bset[5];
                                de.electrical_shock = bset[4];
                                de.motor_encoder    = bset[3];
                                de.overheating      = bset[2];
                                de.input_voltage    = bset[0];
                                vis( value_tag{}, de );
                        },
                        [&]( error_tag, auto e ) {
                                vis( error_tag{}, e );
                        } ) );
        }
};

}  // namespace schpin
