#include "fw/base.h"
#include "fw/lewan/config.h"
#include "fw/lewan/context.h"
#include "protocol.h"

#include <emlabcpp/pid.h>
#include <schpin_lib/algorithm.h>
#include <schpin_lib/logging.h>
#include <variant>

#pragma once

namespace schpin
{

struct lewan_goto_event
{
        lewan_angle g;
        lewan_time  t;
};

struct lewan_spin_event
{
        lewan_angular_velocity av;
        lewan_time             t;
};

struct lewan_spin_until_event
{
        lewan_angle            g;
        lewan_angular_velocity av;
};

struct lewan_config_event
{
        master_servo_config_variant var;
};

using lewan_servo_events_variant = std::variant<  //
    lewan_goto_event,
    lewan_spin_event,
    lewan_spin_until_event,
    lewan_config_event >;

class lewan_servo
{
        struct activate_task
        {
                std::size_t i = 0;
                cmd_roulette< LEWAN_SERVO_ANGLE_OFFSET_READ,
                              LEWAN_SERVO_ANGLE_LIMIT_READ,
                              LEWAN_SERVO_VIN_LIMIT_READ,
                              LEWAN_SERVO_TEMP_MAX_LIMIT_READ,
                              LEWAN_SERVO_OR_MOTOR_MODE_READ,
                              LEWAN_SERVO_LED_CTRL_READ,
                              LEWAN_SERVO_POS_READ >
                    startup_roulette;
        };

        struct goto_task
        {
                lewan_full_angle start;
                lewan_angle      goal;
                clk_time         start_time;
                lewan_time       time;
                em::pid          vel_pid;
        };

        struct spin_task
        {
                clk_time clock_end;
        };

        struct spin_until_task
        {
                lewan_angle            goal;
                lewan_angular_velocity av;
                int                    pos_diff;
                clk_time               task_start;
        };

        struct idle_task
        {
        };

        using task_variant =
            std::variant< activate_task, goto_task, spin_task, idle_task, spin_until_task >;

        prot_config< servo_config >           config_;
        lewan_context                         context_;
        task_variant                          task_ = { activate_task{} };
        lewan_servo_id                        id_;
        lewan_full_angle                      last_report_;
        message_catcher< lewan_master_group > lewan_catcher_;
        std::size_t                           poll_i_         = 0;
        s_ctl_id                              msg_id_counter_ = 0;

      public:
        using receive_type   = lewan_master_variant;
        using events_variant = lewan_servo_events_variant;

        lewan_servo( lewan_servo_id id )
          : id_( id )
        {
        }

        servo_id get_id() const
        {
                return *id_;
        }

        bool has_task() const
        {
                return !std::holds_alternative< idle_task >( task_ );
        }

        bool is_activating() const
        {
                return std::holds_alternative< activate_task >( task_ );
        }

        const lewan_servo_state& get_state() const
        {
                return *context_;
        }

        template < typename Visitor >
        void stop( Visitor&& vis )
        {
                // TODO: DO NOT DO THIS IF ACTIVATING?
                task_ = idle_task{};
                transmit_servo< LEWAN_SERVO_MOVE_STOP, reply::NO >( vis );
        }

        std::optional< servo_master_message > on_servo_receive( clk_time clock_time,
                                                                s_ctl_id,
                                                                const lewan_master_variant& var )
        {
                lewan_catcher_.check( var );
                context_.on_servo_receive( clock_time, var );

                using pos_tuple = std::tuple< cmd_tag< LEWAN_SERVO_POS_READ >, lewan_full_angle >;
                if ( !std::holds_alternative< pos_tuple >( var ) ) {
                        return {};
                }

                if ( config_->angle_change_report == lewan_angle::get< 0 >() ) {
                        return {};
                }

                if ( angle_dist( last_report_, context_->pos ) < *config_->angle_change_report ) {
                        return {};
                }
                last_report_ = context_->pos;
                return servo_master_group::create< POSITION >( conv( context_->pos ) );
        }

        template < typename Visitor >
        void tick( clk_time clock_time, Visitor&& vis )
        {
                task_ << match(
                    [&]( idle_task ) {},
                    [&]( const activate_task& s ) {
                            activate_tick( s, vis );
                    },
                    [&]( goto_task t ) {
                            // stuff may overflow badyl here
                            float goal_f  = static_cast< float >( *t.goal );
                            float start_f = static_cast< float >( *t.start );
                            float actual_pos =
                                em::map_range( static_cast< float >(
                                                   em::min( *clock_time - *t.start_time,
                                                            static_cast< unsigned >( *t.time ) ) ),
                                               0.f,
                                               static_cast< float >( *t.time ),
                                               start_f,
                                               goal_f );

                            t.vel_pid.update( clock_time.as_float(),
                                              static_cast< float >( *context_->pos ),
                                              actual_pos );

                            transmit_servo< LEWAN_SERVO_OR_MOTOR_MODE_WRITE, reply::NO >(
                                vis,
                                lewan_motor_mode,
                                uint8_t{ 0 },
                                *lewan_angular_velocity::make(
                                    static_cast< int16_t >( t.vel_pid.get_output() ) ) );
                            task_ = t;
                            if ( *t.start_time + *t.time > *clock_time ) {
                                    return;
                            }
                            // random number - meh TODO
                            if ( angle_dist( context_->pos, t.goal ) < 5 ) {
                                    transmit_servo< LEWAN_SERVO_OR_MOTOR_MODE_WRITE, reply::NO >(
                                        vis,
                                        lewan_position_mode,
                                        uint8_t{ 0 },
                                        lewan_angular_velocity::get< 0 >() );
                                    transmit_servo< LEWAN_SERVO_MOVE_TIME_WRITE, reply::NO >(
                                        vis, t.goal, lewan_time::get< 0 >() );
                                    task_ = idle_task{};
                            }
                    },
                    [&]( const spin_task& t ) {
                            if ( t.clock_end > clock_time ) {
                                    return;
                            }
                            task_ = idle_task{};
                            transmit_servo< LEWAN_SERVO_OR_MOTOR_MODE_WRITE, reply::NO >(
                                vis,
                                lewan_position_mode,
                                uint8_t{ 0 },
                                lewan_angular_velocity::get< 0 >() );
                    },
                    [&]( spin_until_task t ) {
                            lewan_full_angle pos  = context_->pos;
                            auto             diff = angle_dist( t.goal, pos );
                            bool             in_range =
                                diff < 5;  // TODO: same random number as in position task
                            bool overrun    = em::sign( t.pos_diff ) != em::sign( diff );
                            bool is_stalled = context_.get_hist_sum() < *config_->angle_stall_range;
                            if ( clock_time - t.task_start < clk_time::from_float( 1.f ) ) {
                                    is_stalled = false;
                            }
                            // TODO: report error if stalled
                            if ( !in_range && !overrun && !is_stalled ) {
                                    task_ = spin_until_task{ t.goal, t.av, diff, t.task_start };
                                    return;
                            }
                            transmit_servo< LEWAN_SERVO_OR_MOTOR_MODE_WRITE, reply::NO >(
                                vis,
                                lewan_position_mode,
                                uint8_t{ 0 },
                                lewan_angular_velocity::get< 0 >() );
                            task_ = idle_task{};
                            if ( !is_stalled ) {
                                    transmit_servo< LEWAN_SERVO_MOVE_TIME_WRITE, reply::NO >(
                                        vis, t.goal, lewan_time::get< 0 >() );
                            }
                    } );
        }

        template < typename Visitor >
        void on_event( ctl_id,
                       clk_time                          clock_time,
                       const lewan_servo_events_variant& var,
                       Visitor&&                         vis )
        {
                var << match(
                    [&]( lewan_goto_event e ) {
                            task_ = goto_task{ context_->pos,
                                               e.g,
                                               clock_time,
                                               e.t,
                                               em::pid::config{ config_->goto_p_gain.as_float(),
                                                                config_->goto_i_gain.as_float(),
                                                                config_->goto_d_gain.as_float(),
                                                                -1000.f,
                                                                1000.f } };
                            transmit_servo< LEWAN_SERVO_OR_MOTOR_MODE_WRITE, reply::NO >(
                                vis,
                                lewan_position_mode,
                                uint8_t{ 0 },
                                lewan_angular_velocity::get< 0 >() );
                    },
                    [&]( lewan_spin_event e ) {
                            task_ = spin_task{ clock_time + clk_time{ *e.t } };
                            // TODO: spot ugly thing
                            transmit_servo< LEWAN_SERVO_OR_MOTOR_MODE_WRITE, reply::NO >(
                                vis, lewan_motor_mode, uint8_t{ 0 }, e.av );
                    },
                    [&]( lewan_spin_until_event e ) {
                            int diff = angle_dist( e.g, context_->pos );
                            // TODO: hardcoded constant
                            if ( diff < 5 ) {
                                    transmit_servo< LEWAN_SERVO_MOVE_TIME_WRITE, reply::NO >(
                                        vis, e.g, lewan_time::get< 0 >() );
                                    return;
                            }
                            task_ = spin_until_task{ e.g, e.av, diff, clock_time };
                            transmit_servo< LEWAN_SERVO_OR_MOTOR_MODE_WRITE, reply::NO >(
                                vis, lewan_motor_mode, uint8_t{ 0 }, e.av );
                    },
                    [&]( lewan_config_event e ) {  //
                            config_.on_master_config_msg(
                                e.var, [&]( servo_master_config_message m ) {
                                        servo_reply_ctl< CONFIG >( vis, m );
                                } );
                    } );
        }

        template < typename Visitor >
        void poll( clk_time, std::size_t i_offset, Visitor&& vis )
        {
                poll_i_ += 1 % 13;
                std::size_t i = ( i_offset + poll_i_ ) % 13;

                switch ( i ) {
                        case 4:
                                transmit_servo< LEWAN_SERVO_VIN_READ, reply::YES >( vis );
                                break;
                        case 8:
                                transmit_servo< LEWAN_SERVO_TEMP_READ, reply::YES >( vis );
                                break;
                        default:
                                transmit_servo< LEWAN_SERVO_POS_READ, reply::YES >( vis );
                                break;
                }
        }

        em::either< lewan_servo_events_variant, servo_master_message >
        on_servo_packet( ctl_id id, master_servo_command_variant subcmd )
        {
                em::either< lewan_servo_events_variant, servo_master_message > res =
                    servo_master_message{};

                subcmd << tuple_match(
                    [&]( cmd_tag< GOTO >, bangle a, btime t ) {
                            res = on_converted< GOTO >(
                                id,
                                a,
                                t,
                                [&]( lewan_angle an, lewan_time ti ) -> lewan_servo_events_variant {
                                        SCHPIN_INFO_LOG( "wololo",
                                                         "s: " << id_ << " goto: " << a << " - "
                                                               << an );
                                        return lewan_goto_event{ an, ti };
                                } );
                    },
                    [&]( cmd_tag< SPIN >, bangular_velocity av, btime t ) {
                            res = on_converted< SPIN >(
                                id,
                                av,
                                t,
                                [&]( lewan_angular_velocity anv,
                                     lewan_time             ti ) -> lewan_servo_events_variant {
                                        return lewan_spin_event{ anv, ti };
                                } );
                    },
                    [&]( cmd_tag< SPIN_UNTIL >, bangle a, bangular_velocity av ) {
                            res = on_converted< SPIN_UNTIL >(
                                id,
                                a,
                                av,
                                [&]( lewan_angle            an,
                                     lewan_angular_velocity anv ) -> lewan_servo_events_variant {
                                        return lewan_spin_until_event{ an, anv };
                                } );
                    },
                    [&]( cmd_tag< POSITION > ) {  //
                            res = make_servo_message< POSITION >( conv( context_->pos ) );
                    },
                    [&]( cmd_tag< TEMP > ) {  //
                            res = make_servo_message< TEMP >( conv( context_->temperature ) );
                    },
                    [&]( cmd_tag< VOLTAGE > ) {  //
                            res = make_servo_message< VOLTAGE >( conv( context_->voltage ) );
                    },
                    [&]( cmd_tag< VELOCITY > ) {  //
                            res = make_servo_message< VELOCITY >( conv( context_->velocity ) );
                    },
                    [&]( cmd_tag< TASK > ) {  //
                            servo_task_enum e = task_ << match(
                                                    [&]( idle_task ) {
                                                            return TASK_IDLE;
                                                    },
                                                    [&]( activate_task ) {
                                                            return TASK_ACTIVATE;
                                                    },
                                                    [&]( goto_task ) {
                                                            return TASK_GOTO;
                                                    },
                                                    [&]( spin_task ) {
                                                            return TASK_SPIN;
                                                    },
                                                    [&]( spin_until_task ) {
                                                            return TASK_SPIN_UNTIL;
                                                    } );

                            res = make_servo_message< TASK >( e );
                    },
                    [&]( cmd_tag< CONFIG >,
                         master_servo_config_variant subvar ) {  //
                            res = lewan_config_event{ subvar };
                    } );

                return res;
        }

      private:
        template < typename Visitor >
        void activate_tick( activate_task s, Visitor&& v )
        {
                switch ( s.i ) {
                        case 0:
                                s.startup_roulette.tick( [&]( auto tag ) {  //
                                        transmit_servo< decltype( tag )::value, reply::YES >( v );
                                } );
                                if ( s.startup_roulette.get_index() != 0 ) {
                                        task_ = s;
                                        return;
                                }
                                break;
                        case 1:
                                lewan_catcher_.set_trap( LEWAN_SERVO_POS_READ );
                                transmit_servo< LEWAN_SERVO_POS_READ, reply::YES >( v );
                                break;
                        case 2:
                                if ( !lewan_catcher_.catched() ) {
                                        return;
                                }
                                break;
                        default:
                                v( activate_tag{}, *id_ );
                                task_ = idle_task{};
                                return;
                }
                s.i += 1;
                task_ = s;
        }

        template < messages_enum ID, typename LT, typename RT, typename Function >
        em::either< lewan_servo_events_variant, servo_master_message >
        on_converted( ctl_id cid, LT lh, RT rh, Function&& f )
        {
                auto opt_lh = conv( lh );
                auto opt_rh = conv( rh );
                if ( !opt_lh ) {
                        return make_servo_error_message< INPUT_INVALID_VALUE_E >(
                            cid, ID, uint32_t{ 0 } );
                }
                if ( !opt_rh ) {
                        return make_servo_error_message< INPUT_INVALID_VALUE_E >(
                            cid, ID, uint32_t{ 1 } );
                }
                return f( *opt_lh, *opt_rh );
        }

        std::optional< lewan_angle > conv( bangle a )
        {
                float mapped = map_val< float >( a,
                                                 config_->input_low,
                                                 config_->input_high,
                                                 *config_->servo_low,
                                                 *config_->servo_high );
                return lewan_angle::make( static_cast< uint16_t >( mapped ) );
        }

        std::optional< lewan_time > conv( btime t )
        {
                return lewan_time::make( static_cast< uint16_t >( t.as_float() * 1000.f ) );
        }

        std::optional< lewan_angular_velocity > conv( bangular_velocity av )
        {
                float s_diff = static_cast< float >( *config_->servo_high ) -
                               static_cast< float >( *config_->servo_low );
                float i_diff = config_->input_high.as_float() - config_->input_low.as_float();
                float coeff  = std::fabs( s_diff ) / std::fabs( i_diff );

                float res = av.as_float() * coeff;

                if ( em::sign( s_diff ) != em::sign( i_diff ) ) {
                        res *= -1;
                }
                return lewan_angular_velocity::make( static_cast< int16_t >( res ) );
        }

        bangle conv( lewan_full_angle a )
        {
                auto res = bangle::from_float(
                    map_val< float >( static_cast< float >( *a ),
                                      static_cast< float >( *config_->servo_low ),
                                      static_cast< float >( *config_->servo_high ),
                                      config_->input_low.as_float(),
                                      config_->input_high.as_float() ) );
                return res;
        }

        bangle conv( lewan_angle a )
        {
                return conv( lewan_full_angle{ a } );
        }

        btemp conv( lewan_temp temp )
        {
                return btemp::from_float( *temp );
        }

        bvoltage conv( lewan_voltage volt )
        {
                return bvoltage{ *volt };
        }

        bangular_velocity conv( lewan_angular_velocity vel )
        {
                return bangular_velocity::from_float(
                    map_val< float >( *vel,
                                      int16_t( *config_->servo_low ),
                                      int16_t( *config_->servo_high ),
                                      config_->input_low,
                                      config_->input_high ) );
        }

        template < messages_enum ID, typename Visitor, typename... Args >
        void servo_reply_ctl( Visitor&& v, Args&&... args )
        {
                v( control_tag{}, id_, servo_master_group::create< ID >( args... ) );
        }

        template < error_protocol_enum ID, typename Visitor, typename... Args >
        void servo_reply_ctl_error( ctl_id cid, Visitor&& v, Args&&... args )
        {
                servo_reply_ctl< ERROR >( v, cid, servo_error_group::create< ID >( args... ) );
        }

        template < lewan_message_enum ID, reply Rpl, typename Visitor, typename... Args >
        void transmit_servo( Visitor&& v, Args... args )
        {
                msg_id_counter_++;
                v( servo_tag{},
                   msg_id_counter_,
                   Rpl,
                   master_lewan_protocol::create( id_,
                                                  master_lewan_group::create< ID >( args... ) ) );
        }

        template < error_protocol_enum ID, typename... Args >
        servo_master_message make_servo_error_message( ctl_id id, Args&&... args )
        {
                return make_servo_message< ERROR >( id,
                                                    servo_error_group::create< ID >( args... ) );
        }

        template < messages_enum ID, typename... Args >
        servo_master_message make_servo_message( Args&&... args )
        {
                auto msg = servo_master_group::create< ID >( args... );
                return msg;
        }
};

}  // namespace schpin
