#include "node/leg/ik.h"

#include <schpin_lib/model/leg.h>
#include <schpin_lib/model/leg/fk.h>
#include <schpin_lib/model/leg/storage.h>
#include <schpin_lib/srdf.h>

#pragma once

namespace schpin
{

inline sized_array< const joint*, dynamic_size > make_revolute_mapping( const joint& root )
{
        std::vector< const joint* > res;

        std::function< void( const joint& ) > f = [&]( const joint& joint ) {
                if ( std::holds_alternative< revolute_joint >( joint.jclass ) ) {
                        res.push_back( &joint );
                }
                if ( joint.sub_link.child_ptr ) {
                        f( *joint.sub_link.child_ptr );
                }
        };

        f( root );

        return sized_array< const joint*, dynamic_size >{ res };
}

inline pose< robot_frame > first_revolute_offset( const joint& root )
{
        if ( std::holds_alternative< revolute_joint >( root.jclass ) ) {
                return root.offset;
        }

        if ( root.sub_link.child_ptr ) {
                return transform( first_revolute_offset( *root.sub_link.child_ptr ), root.offset );
        }

        return pose< robot_frame >{};
}

template < typename T >
using dof_array = sized_array< T, dynamic_size >;

using leg_model_error = error< struct leg_model_error_tag >;

class leg_model
{

        std::string               name_;
        std::unique_ptr< joint >  root_joint_ptr_;
        dof_array< const joint* > revolute_mapping_;
        fk_solver< dynamic_size > fk_;
        ik_solver                 ik_;

      public:
        leg_model( std::string name, joint jnt )
          : name_( std::move( name ) )
          , root_joint_ptr_( std::make_unique< joint >( std::move( jnt ) ) )
          , revolute_mapping_( make_revolute_mapping( *root_joint_ptr_ ) )
          , fk_( first_revolute_offset( *root_joint_ptr_ ), revolute_mapping_ )
          , ik_( first_revolute_offset( *root_joint_ptr_ ), revolute_mapping_ )
        {
        }

        const std::string& get_name() const
        {
                return name_;
        }

        const dof_array< const joint* >& get_revolute_mapping() const
        {
                return revolute_mapping_;
        }

        opt_error< leg_model_error >
        check_angles_range( const dof_array< em::angle >& angles ) const
        {
                std::vector< leg_model_error > tmp_errors = filter_optionals(
                    em::zip( angles, revolute_mapping_ ),
                    em::uncurry |
                        [&]( em::angle    alpha,
                             const joint* joint_ptr ) -> std::optional< leg_model_error > {
                            SCHPIN_ASSERT(
                                std::holds_alternative< revolute_joint >( joint_ptr->jclass ) );

                            const revolute_joint& rev_joint =
                                std::get< revolute_joint >( joint_ptr->jclass );

                            bool out_of_range =
                                alpha < rev_joint.min_angle || alpha > rev_joint.max_angle;

                            if ( out_of_range ) {
                                    return { E( leg_model_error )
                                             << "Angle " << alpha << " for joint "
                                             << joint_ptr->name
                                             << " is out of range: " << rev_joint.min_angle << " - "
                                             << rev_joint.max_angle };
                            }
                            return {};
                    } );

                if ( tmp_errors.empty() ) {
                        return {};
                }
                return { E( leg_model_error ).group_attach( tmp_errors )
                         << " Problem with leg: " << name_ << " at angles: " << angles };
        }

        em::either< dof_array< em::angle >, ik_error >
        get_angles( const point< 3, robot_frame >& target,
                    const dof_array< em::angle >&  origin,
                    std::size_t                    stop_steps = 32 ) const
        {
                // TODO: ugly constnat
                em::distance stop_error{ 0.000005f };
                return ik_.calculate_angles( fk_, origin, target, stop_error, stop_steps );
        }

        point< 3, robot_frame > get_tip_position( const dof_array< em::angle >& angles ) const
        {
                return fk_.get_tip_position( angles );
        }

        const joint& get_root_joint() const
        {
                return *root_joint_ptr_;
        }

        const ik_solver& get_ik() const
        {
                return ik_;
        }

        const fk_solver< dynamic_size >& get_fk() const
        {
                return fk_;
        }
};

inline std::vector< em::either< leg_model, load_error > >
load_leg_models( const urdf::Model& urdf_model, const srdf_information& srdf_info )
{
        urdf::LinkConstSharedPtr model_root = urdf_model.getLink( srdf_info.body_link );

        return map_f_to_v(  //
            srdf_info.root_joints,
            [&]( std::string root_joint_name ) -> em::either< leg_model, load_error > {
                    auto root_ptr = urdf_model.getJoint( root_joint_name );
                    if ( !root_ptr ) {
                            return { E( load_error ) << "Root joint " << root_joint_name
                                                     << " is missing in the model. " };
                    }
                    return parse_joint_rec( urdf_model, root_ptr )
                        .convert_right( [&]( const urdf_parse_error& e ) {
                                return E( load_error ).attach( e ) << "Error during the parsing "
                                                                      "of legs for robot";
                        } )
                        .convert_left( [&]( joint root_joint ) {
                                return leg_model{ root_joint_name, std::move( root_joint ) };
                        } );
            } );
}

inline std::ostream& operator<<( std::ostream& os, const leg_model& l )
{
        return os << l.get_root_joint() << "\n"
                  << "ik:\n"
                  << l.get_ik();
}

}  // namespace schpin
