#include "schpin_lib/algorithm.h"
#include "schpin_lib/error.h"
#include "schpin_lib/print.h"
#include "schpin_lib/util.h"
#include "schpin_lib/view.h"
#include "schpin_lib_exp/algorithm.h"
#include "schpin_lib_exp/bounded.h"

#include <readline/history.h>
#include <readline/readline.h>
#include <variant>

#pragma once

namespace schpin
{
namespace type_console
{

        struct load_commands_event
        {
                std::string filename;
        };

        using color               = term_color;
        using user_input          = em::view< std::vector< std::string >::const_iterator >;
        using user_input_iterator = typename user_input::iterator;

        struct match_result
        {
                std::vector< std::string >                  state{};
                std::optional< std::vector< std::string > > opt_matched_cmd{};

                match_result() = default;

                match_result( const user_input&                           input,
                              std::optional< std::vector< std::string > > opt_line = {} )
                  : state( input.begin(), input.end() )
                  , opt_matched_cmd( std::move( opt_line ) )
                {
                }
        };

        using console_error = std::string;

        struct no_event
        {
        };

        struct completion_context
        {
                const std::string                              key;
                const std::vector< const completion_context* > children;

                completion_context( std::string                              key_i,
                                    std::vector< const completion_context* > children_i )
                  : key( std::move( key_i ) )
                  , children( std::move( children_i ) )
                {
                }
        };

        struct completion_candidates
        {
                std::string                     prefix;
                std::vector< std::string_view > candidates{};
        };

        template < typename View >
        completion_candidates find_candidates( View words, const completion_context* context )
        {
                if ( context == nullptr ) {
                        return { std::string{ words.front() } };
                }
                if ( words.empty() ) {
                        std::vector< std::string_view > candidates =
                            map_f_to_v( context->children,
                                        [&]( const completion_context* child ) -> std::string_view {
                                                return child->key;
                                        } );
                        return completion_candidates{ "", candidates };
                }
                auto child_iter =
                    em::find_if( context->children, [&]( const completion_context* child ) {  //
                            return child->key == words.front();
                    } );
                if ( child_iter != context->children.end() ) {
                        completion_candidates sub =
                            find_candidates( em::tail( words ), *child_iter );
                        return { ( *child_iter )->key + " " + sub.prefix, sub.candidates };
                }
                std::vector< std::string_view > filtered =
                    filter_optionals( context->children, [&]( const completion_context* child ) {
                            if ( child->key.compare( 0, words.front().size(), words.front() ) ==
                                 0 ) {
                                    return std::make_optional( std::string_view{ child->key } );
                            }
                            return std::optional< std::string_view >{};
                    } );
                return { "", filtered };
        }

        struct console_cmd_context
        {
                const std::vector< std::string >            keywords;
                const color                                 cmd_color;
                const std::string                           merged_keywords;
                const std::shared_ptr< completion_context > completion_ptr;

                console_cmd_context(
                    std::vector< std::string >                      keywords_i,
                    color                                           color_i,
                    const std::vector< const completion_context* >& compl_child = {} )
                  : keywords( std::move( keywords_i ) )
                  , cmd_color( color_i )
                  , merged_keywords( em::joined( keywords, std::string{ ", " } ) )
                  , completion_ptr(
                        std::make_shared< completion_context >( keywords.at( 0 ), compl_child ) )
                {
                }
        };

        template < typename Word >
        bool keyword_matches( const Word& word, const console_cmd_context& context )
        {
                auto iter = em::find_if( context.keywords, [&]( const std::string& key ) {  //
                        return key == word;
                } );
                return iter != context.keywords.end();
        }

        std::vector< std::string > prompt( const std::vector< std::string >& p,
                                           const completion_context*         root );

        struct help_command
        {
                using event_type = no_event;
                console_cmd_context              context{ { "help", "h" }, color::CYAN };
                const std::vector< std::string > description{ "prints help of this level" };

                template < typename ParentLevel >
                match_result match( const user_input&          input,
                                    const user_input_iterator& item,
                                    const ParentLevel&         parent ) const
                {
                        if ( std::next( item ) != input.end() ) {
                                std::cout << "Help does not take any arguments\n";
                                return match_result{ em::init( input ) };
                        }
                        std::cout << std::endl;  // hmmmm ?
                        std::vector< std::string > keywords;

                        std::size_t indent =
                            em::max_elem( parent.items, [&]( const auto& item ) {  //
                                    return item.context.merged_keywords.size();
                            } );
                        std::cout << std::left;

                        em::for_each( parent.items, [&]( const auto& item ) {
                                std::cout << "    " << item.context.cmd_color
                                          << std::setw( static_cast< int >( indent ) )
                                          << item.context.merged_keywords << color::RESET << " - "
                                          << item.description.front() << std::endl;

                                em::for_each(
                                    em::tail( item.description ), [&]( const std::string& line ) {
                                            std::cout << "    "
                                                      << std::setw( static_cast< int >( indent ) )
                                                      << " "
                                                      << "   " << line << std::endl;
                                    } );
                        } );

                        std::cout << std::endl;
                        return match_result{ em::init( input ) };
                }

                em::either< no_event, console_error > parse( const user_input&,
                                                             const user_input_iterator& ) const
                {
                        return { console_error{ "can't parse help command" } };
                }
        };

        struct parent_command
        {
                using event_type = no_event;
                console_cmd_context              context{ { ".." }, color::CYAN };
                const std::vector< std::string > description{ "goes up in the control level" };

                template < typename ParentLevel >
                match_result match( const user_input& input,
                                    const user_input_iterator&,
                                    const ParentLevel& ) const
                {
                        if ( input.size() == 1 ) {

                                return match_result{ em::init( input ) };
                        }
                        return match_result{ em::init( em::init( input ) ) };
                }

                em::either< no_event, console_error > parse( const user_input&,
                                                             const user_input_iterator& ) const
                {
                        return { console_error{ "can't parse parent command" } };
                }
        };

        struct exit_command
        {
                using event_type = no_event;
                console_cmd_context              context{ { "quit", "q", "exit" }, color::CYAN };
                const std::vector< std::string > description{ "exits the entire app" };

                template < typename ParentLevel >
                match_result match( const user_input& input,
                                    const user_input_iterator&,
                                    const ParentLevel& ) const
                {
                        std::exit( 0 );
                        return { input };
                }

                em::either< no_event, console_error > parse( const user_input&,
                                                             const user_input_iterator& ) const
                {
                        return { console_error{ "can't parse console exit command" } };
                }
        };

        template < typename EventT, typename... Ts >
        struct level_impl
        {
                using event_type = EventT;

                std::string                      name;
                const std::vector< std::string > description;
                std::tuple< Ts... >              items;
                console_cmd_context              context;

                level_impl( std::string name_i, std::string description_i, Ts... items_i )
                  : name( std::move( name_i ) )
                  , description( { std::move( description_i ) } )
                  , items( std::move( items_i )... )
                  , context(
                        { name },
                        color::MAGENTA,
                        map_f_to_v( items, []( const auto& item ) -> const completion_context* {
                                return item.context.completion_ptr.get();
                        } ) )
                {
                }

                template < typename NewEvent, typename T >
                level_impl< NewEvent, T, Ts... > extend( T&& item ) &&
                {
                        return std::apply(
                            [&]( auto&&... items ) {
                                    return level_impl< NewEvent, T, Ts... >{
                                        name,
                                        description.front(),
                                        std::forward< T >( item ),
                                        std::move( items )... };
                            },
                            std::move( items ) );
                }

                template < typename ParentLevel >
                match_result match( const user_input&   input,
                                    user_input_iterator input_cmd_iter,
                                    const ParentLevel& ) const
                {
                        input_cmd_iter++;
                        if ( input_cmd_iter == input.end() ) {
                                return { input };
                        }
                        std::optional< match_result > opt_child_res = first_optional(
                            items, [&]( const auto& item ) -> std::optional< match_result > {
                                    if ( !keyword_matches( *input_cmd_iter, item.context ) ) {
                                            return {};
                                    }

                                    return item.match( input, input_cmd_iter, *this );
                            } );

                        if ( opt_child_res ) {
                                return *opt_child_res;
                        }

                        std::cerr << "failed to match command on this level " << std::endl;
                        return match_result{ em::view{ input.begin(), input_cmd_iter } };
                }

                em::either< event_type, console_error >
                parse( const user_input& input, user_input_iterator input_cmd_iter ) const
                {
                        input_cmd_iter++;
                        if ( input_cmd_iter == input.end() ) {
                                return { console_error{ "can't do level change in parser" } };
                        }

                        std::optional< em::either< event_type, console_error > > opt_child_res =
                            first_optional(
                                items,
                                [&]( const auto& item )
                                    -> std::optional< em::either< event_type, console_error > > {
                                        using item_event =
                                            typename std::decay_t< decltype( item ) >::event_type;
                                        if ( !keyword_matches( *input_cmd_iter, item.context ) ) {
                                                return {};
                                        }

                                        if constexpr ( std::is_same_v< item_event, no_event > ) {
                                                return {};
                                        } else {
                                                return item.parse( input, input_cmd_iter )
                                                    .convert_left(
                                                        [&]( const auto& sub_event ) {  //
                                                                return event_type{ sub_event };
                                                        } );
                                        }
                                } );
                        if ( opt_child_res ) {
                                return *opt_child_res;
                        }
                        return { console_error{ "failed to find command on this level" } };
                }
        };  // namespace schpin

        template < typename RootEventT, typename... Ts >
        class console_impl
        {
              public:
                using root_event = RootEventT;
                using root_level = level_impl< root_event, Ts... >;

              private:
                root_level root_;

              public:
                console_impl( Ts... items )
                  : root_( "", "", std::move( items )... )
                {
                }

                console_impl( root_level&& level )
                  : root_( std::move( level ) )
                {
                }

                template < typename NewRootEvent, typename T >
                console_impl< NewRootEvent, T, Ts... > extend( T&& item ) &&
                {
                        return { std::move( root_ ).template extend< NewRootEvent >(
                            std::forward< T >( item ) ) };
                }

                template < typename UnaryFunction >
                void match_stream( std::istream& ifile, UnaryFunction&& f )
                {
                        std::string line;
                        while ( std::getline( ifile, line ) ) {
                                std::vector< std::string > user_input =
                                    map_f_to_v( split( line ), [&]( std::string_view view ) {  //
                                            return std::string{ view };
                                    } );
                                match_result res = match( user_input );
                                if ( res.opt_matched_cmd ) {
                                        f( *res.opt_matched_cmd );
                                }
                        }
                }

                template < typename UnaryFunction >
                void infinitely_match( UnaryFunction f )
                {
                        match_result result;
                        while ( true ) {
                                result = prompt_match( result.state );
                                if ( result.opt_matched_cmd ) {
                                        f( *result.opt_matched_cmd );
                                }
                        }
                }

                match_result prompt_match( const std::vector< std::string >& prev_path ) const
                {
                        std::vector< std::string > user_input =
                            prompt( prev_path, root_.context.completion_ptr.get() );
                        return match( merge( prev_path, user_input ) );
                }

                template < typename Input >
                match_result match( const Input& input ) const
                {
                        auto iview = em::view{ input };
                        return root_.match( iview, std::prev( iview.begin() ), *this );
                }

                template < typename Input >
                em::either< root_event, console_error > parse( const Input& input ) const
                {
                        auto iview = em::view{ input };
                        return root_.parse( iview, std::prev( iview.begin() ) );
                }
        };

        template < typename RootEvent, typename... Ts >
        console_impl< RootEvent, Ts..., help_command, exit_command > console( Ts... items )
        {
                return { std::move( items )..., help_command{}, exit_command{} };
        }

        template < typename ResultType, typename... Ts >
        level_impl< ResultType, Ts..., help_command, parent_command, exit_command >
        level( const std::string& name, const std::string& description, Ts... items )
        {
                return { name,
                         description,
                         std::move( items )...,
                         help_command{},
                         parent_command{},
                         exit_command{} };
        }

        template < typename T >
        struct arg_parser
        {
                static em::either< user_input_iterator, console_error >
                match( const user_input& input )
                {
                        if ( input.empty() ) {
                                return { "Missing argument." };
                        }
                        return { std::next( input.begin() ) };
                }
                static em::either< T, console_error > parse( const user_input& input )
                {
                        T                 tmp;
                        std::stringstream ss{ input.front() };
                        ss >> tmp;
                        return { tmp };
                }
        };

        template <>
        struct arg_parser< uint8_t >
        {
                static em::either< user_input_iterator, console_error >
                match( const user_input& input )
                {
                        return arg_parser< unsigned >::match( input );
                }
                static em::either< uint8_t, console_error > parse( const user_input& input )
                {
                        return arg_parser< unsigned >::parse( input ).convert_left(
                            [&]( unsigned val ) {  //
                                    return static_cast< uint8_t >( val );
                            } );
                }
        };

        template < typename T, T MinVal, T MaxVal >
        struct arg_parser< bounded< T, MinVal, MaxVal > >
        {
                static em::either< user_input_iterator, console_error >
                match( const user_input& input )
                {
                        if ( input.empty() ) {
                                return { "Missing argument." };
                        }
                        return { std::next( input.begin() ) };
                }

                static em::either< bounded< T, MinVal, MaxVal >, console_error >
                parse( const user_input& input )
                {
                        return arg_parser< T >::parse( input ).bind_left(
                            [&]( T val )
                                -> em::either< bounded< T, MinVal, MaxVal >, console_error > {
                                    auto opt = bounded< T, MinVal, MaxVal >::make( val );
                                    if ( opt ) {
                                            return { *opt };
                                    }
                                    return { console_error{ "Value is out of bounds." } };
                            } );
                }
        };

        template < typename T >
        struct arg_parser< std::optional< T > >
        {

                static em::either< user_input_iterator, console_error >
                match( const user_input& input )
                {
                        if ( input.empty() ) {
                                return { input.end() };
                        }
                        return arg_parser< T >::match( input );
                }

                static em::either< std::optional< T >, console_error >
                parse( const user_input& input )
                {
                        if ( input.empty() ) {
                                return { std::optional< T >{} };
                        }

                        return arg_parser< T >::parse( input ).convert_left( [&]( T t ) {  //
                                return std::make_optional< T >( t );
                        } );
                }
        };

        template < typename T >
        struct arg
        {
                std::string name;
                std::string description;

                using type = T;

                arg( std::string name_i, std::string description_i )
                  : name( std::move( name_i ) )
                  , description( std::move( description_i ) )
                {
                }

                std::string get_type_name() const
                {
                        char const* type_name = typeid( T ).name();
                        return boost::core::demangle( type_name );
                }
        };

        template < class T, class Tuple, size_t... Is >
        em::either< T, console_error > construct_from_args( user_input input,
                                                            Tuple&&    args_tuple,
                                                            std::index_sequence< Is... > )
        {
                std::vector< em::either< user_input, console_error > > inputs =
                    map_f_to_v( args_tuple, [&]( const auto& arg ) {
                            using arg_t = typename std::decay_t< decltype( arg ) >::type;
                            return arg_parser< arg_t >::match( input ).convert_left(
                                [&]( user_input_iterator end ) {
                                        user_input actual( input.begin(), end );
                                        input = user_input{ end, input.end() };
                                        return actual;
                                } );
                    } );

                auto partitioned = partition_eithers( inputs );

                if ( !partitioned.right.empty() ) {
                        return { em::joined( partitioned.right, std::string{ "\n" } ) };
                }

                auto f = [&]( const auto& arg, std::size_t i ) {
                        using arg_t = typename std::decay_t< decltype( arg ) >::type;

                        return arg_parser< arg_t >::parse( partitioned.left[i] );
                };

                return assemble_left_collect_right(
                           f( std::get< Is >( std::forward< Tuple >( args_tuple ) ), Is )... )
                    .convert_left( [&]( const auto& t ) {  //
                            return T{ std::get< Is >( t )... };
                    } )
                    .convert_right( [&]( const auto& errors ) {  //
                            return em::joined( errors, std::string{ "\n" } );
                    } );
        }

        template < class T, class Tuple >
        em::either< T, console_error > construct_from_args( const user_input& input,
                                                            Tuple&&           args_tuple )
        {
                if constexpr ( std::tuple_size_v< std::decay_t< Tuple > > != 0 ) {
                        return construct_from_args< T >(
                            input,
                            std::forward< Tuple >( args_tuple ),
                            std::make_index_sequence<
                                std::tuple_size< std::decay_t< Tuple > >::value >{} );
                } else {
                        return T{};
                }
        }

        template < typename EventTypeT, typename... Args >
        struct command_impl
        {
                using event_type = EventTypeT;
                std::string                name;
                console_cmd_context        context;
                std::vector< std::string > description;
                std::tuple< Args... >      args;

                command_impl( std::string name_i, std::string description_line, Args... args_i )
                  : name( std::move( name_i ) )
                  , context( { name }, color::BLUE )
                  , args( args_i... )
                {
                        description.push_back( std::move( description_line ) );

                        if ( std::tuple_size< std::tuple< Args... > >::value != 0 ) {
                                description.push_back( "Arguments: " );
                        }

                        em::for_each( args, [&]( const auto& arg ) {  //
                                std::stringstream ss;
                                ss << "  " << color::CYAN << arg.name << color::RESET
                                   << " - " + arg.description << "(" << arg.get_type_name() << ")";
                                description.push_back( ss.str() );
                        } );
                }

                template < typename ParentLevel >
                match_result match( const user_input&          input,
                                    const user_input_iterator& iter,
                                    const ParentLevel& ) const
                {
                        auto opt_cmd =
                            parse( input, iter )
                                .convert_left( [&]( const event_type& ) {
                                        return std::make_optional( std::vector< std::string >{
                                            input.begin(), input.end() } );
                                } )
                                .convert_right( [&]( const console_error& e )
                                                    -> std::optional< std::vector< std::string > > {
                                        std::cout << e;
                                        return {};
                                } )
                                .join();
                        return match_result{ em::view{ input.begin(), iter }, opt_cmd };
                }

                em::either< event_type, console_error >
                parse( const user_input& input, const user_input_iterator& iter ) const
                {
                        return construct_from_args< event_type >(
                            em::view( std::next( iter ), input.end() ), args );
                }
        };

        template < typename EventType, typename... Args >
        command_impl< EventType, Args... > command( const std::string& name,
                                                    const std::string& description,
                                                    Args... args )
        {
                return { name, description, args... };
        }

        template < typename Tag, typename... Events >
        struct console_event_group
        {
                using variant = std::variant< Events... >;

                variant sub_cmd;

                template < typename T >
                console_event_group( T item )
                  : sub_cmd( std::move( item ) )
                {
                }
        };
}  // namespace type_console
using type_console::arg;
using type_console::console;
using type_console::console_error;
using type_console::console_event_group;
using type_console::level;
using type_console::load_commands_event;
using type_console::match_result;
using type_console::prompt;
}  // namespace schpin
