#include "emlabcpp/zip.h"
#include "schpin_lib_exp/convert_iterator.h"
#include "schpin_lib_exp/overload.h"
#include "schpin_lib_exp/protocol/item.h"
#include "schpin_lib_exp/static_view.h"

#pragma once

namespace schpin
{

template < typename... Items >
struct protocol_tuple
{

        template < typename... NewItems >
        using with_items = protocol_tuple< Items..., protocol_item< NewItems >... >;

        using value_tuple = std::tuple< typename Items::value_type... >;

        static constexpr uint16_t    item_n = static_cast< uint16_t >( sizeof...( Items ) );
        static constexpr std::size_t size   = ( Items::size + ... + 0 );
        static constexpr std::tuple< Items... > items{};
        using message_t  = prot_message< size, protocol_tuple >;
        using value_type = value_tuple;

        static_assert( sizeof...( Items ) < std::numeric_limits< uint16_t >::max(),
                       "Too many items for protocol tuple" );

        static std::string stringify( value_tuple v )
        {
                std::stringstream ss;
                em::for_each(
                    em::zip( v, items ),
                    em::uncurry | [&]( const auto& value, const auto& item ) {  //
                            ss << std::decay_t< decltype( item ) >::stringify( value ) << " | ";
                    } );

                return ss.str();
        }

        static std::string stringify( const message_t& m )
        {
                std::stringstream ss;
                extract( m ).match(
                    [&]( value_tuple v ) {
                            ss << " { " << stringify( std::move( v ) ) << " || "
                               << convert_view< int >( m ) << " } ";
                    },
                    [&]( auto ) {  //
                            ss << convert_view< int >( m );
                    } );
                return ss.str();
        }

        template < typename... Args >
        static constexpr message_t serialize( Args&&... items_val )
        {
                std::array< uint8_t, size > buff = { 0 };
                std::size_t used = serialize_at( make_static_view( buff ), items_val... );
                return message_t{ buff, used };
        }

        template < typename... Args >
        static constexpr std::size_t serialize_at( static_view< size, uint8_t* > target,
                                                   Args&&... items_val )
        {
                return assemble_at( std::make_tuple( std::move( items_val )... ),
                                    target,
                                    std::make_index_sequence< item_n >() );
        }

        template < typename Message, typename... Visitors >
        static std::size_t match( Message&& m, Visitors&&... vis )
        {
                return match( m, schpin::match( vis... ) );
        }

        template < typename Message, typename Visitor >
        static std::size_t match( Message&& m, Visitor&& vis )
        {
                return deserialize(
                    m, std::forward< Visitor >( vis ), std::make_index_sequence< item_n >() );
        }

        template < typename Message >
        static em::either< value_type, protocol_error_record > extract( Message&& m )
        {
                // TODO: the error callback can be called multiple times, maybe allow record notice
                // that?
                em::either< value_tuple, protocol_error_record > res = protocol_error_record{};
                match(
                    m,
                    [&]( error_tag, protocol_error_record rec ) { res = rec; },
                    [&]( auto... items ) { res = std::make_tuple( items... ); } );
                return res;
        }

      private:
        template < typename Tuple, std::size_t... Is >
        static std::size_t assemble_at( Tuple                         values,
                                        static_view< size, uint8_t* > target,
                                        std::index_sequence< Is... > )
        {
                uint8_t* beg = target.begin();
                em::for_each(
                    em::zip( items, values ), em::uncurry | [&]( auto arg_item, auto value ) {
                            auto opt_view =
                                try_static_view< decltype( arg_item )::size >( beg, target.end() );
                            if ( opt_view ) {
                                    beg += decltype( arg_item )::serialize_at( *opt_view, value );
                            } else {
                                    // TODO: make error handling
                            }
                    } );
                return static_cast< std::size_t >( beg - target.begin() );
        }

        template < typename Message, typename Visitor, std::size_t... Is >
        static std::size_t deserialize( Message&& source,
                                        Visitor&& visitor,
                                        std::index_sequence< Is... > )
        {
                std::tuple< std::optional< typename Items::value_type >... > tmp;

                auto beg         = source.begin();
                bool some_failed = false;
                em::for_each(
                    em::zip( items, em::tuple_of_constants_t< Is... >{} ),
                    em::uncurry | [&]( auto tpl_item, auto index ) {
                            using item = decltype( tpl_item );
                            beg += item::deserialize(  //
                                em::view{ beg, source.end() },
                                schpin::match(
                                    [&]( value_tag, auto value ) {
                                            std::get< decltype( index )::value >( tmp ) =
                                                std::make_optional( value );
                                    },
                                    [&]( error_tag, auto err ) {
                                            err.byte_index +=
                                                static_cast< std::size_t >( beg - source.begin() );
                                            visitor( error_tag{}, err );
                                            some_failed = true;
                                    } ) );
                    } );

                if ( !some_failed ) {
                        std::apply(
                            [&]( auto... opt_items ) {  //
                                    visitor( ( *opt_items )... );
                            },
                            tmp );
                }

                return static_cast< std::size_t >( beg - source.begin() );
        }
};

}  // namespace schpin
