
/** Functions used to detect collision between 3D primitives.
 *
 * This is polymorphic code that uses templates. So for each primitive combination checked, separate
 * code is generated.
 *
 * There are two primary functions:
 *   bool collides<A,B>(A,B) - full collision check between primitive of type A and primitive of
 * type B, uses the Separating Axis Test. bool fastCollides<A,B>(A,B) - fast collision check between
 * primitive of type A and primitive of type B, uses min-spheres to detect the collision
 *
 *
 * SAT solution inspired from:
 *   https://gamedevelopment.tutsplus.com/tutorials/collision-detection-using-the-separating-axis-theorem--gamedev-169
 *
 *
 */

#include "collision/comparators.h"
#include "collission_interfaces.h"
#include "schpin_lib/geom/point.h"
#include "util.h"

#pragma once

namespace schpin
{

/* Calculates the smallest and biggest distance of all points from 'T' projected on axis. These are
 * all necessary points for Separation Axis Test. The distances are between the projected points and
 * [0,0,0] coordinate on that axis. The axis is defined only by it's direction.
 *
 * O(n), where 'n' is number of points
 */
template < typename T, std::size_t N >
[[gnu::hot]] constexpr em::min_max< em::distance >
projection_range( const T& t, const vec< N >& axis_direction ) noexcept
{
        return em::min_max_elem( t.get_points(), [&]( const auto& point ) {  //
                return axis_projection_distance( point, axis_direction );
        } );
}

/* Checks if the provided axis detects an separation between 'LH' and 'RH' objects. The axis is
 * defined only by it's direction. All points of each object are projected on same axis and in case
 * they do not overlap, the two objects are not in collision - the separation was detected.
 *
 * O(n+k), where 'n' us number of points for 'LH' and 'k' is number of points for 'RH'.
 */
template < bool CanIntersect, typename LH, typename RH, std::size_t N >
[[gnu::hot]] constexpr bool axis_separates( const LH&       lh,
                                            const RH&       rh,
                                            const vec< N >& axis_direction ) noexcept
{
        em::min_max< em::distance > lh_range   = projection_range( lh, axis_direction );
        em::min_max< em::distance > rh_range   = projection_range( rh, axis_direction );
        bool                        no_overlap = true;
        if constexpr ( CanIntersect ) {
                no_overlap = lh_range.min >= rh_range.max || rh_range.min >= lh_range.max;
        } else {
                no_overlap = lh_range.min > rh_range.max || rh_range.min > lh_range.max;
        }
        // https://stackoverflow.com/questions/3269434/whats-the-most-efficient-way-to-test-two-integer-ranges-for-overlap

        return no_overlap;
}

/* Checks collision between 'LH' and 'RH' objects. The Separation Axis Test is used and only axes of
 * the first object are considered. For each axis, the axisSeparates function is used. The method
 * returns true in case that not a single axis separates the objects. In case at least one axis
 * separates the objects, false is returned.
 *
 * O(a*n), where 'a' is number of axis of 'LH' and 'n' is the sum of points provided by 'LH' and
 * 'RH'
 */
template < typename LH, typename RH, std::size_t N >
[[gnu::hot]] constexpr bool collides_single_axes( const LH& lh, const RH& rh ) noexcept
{
        return em::none_of( lh.get_separation_axes(),
                            [&]( const vec< N >& axis_direction ) {
                                    return axis_separates< false >( lh, rh, axis_direction );
                            } ) &&
               em::none_of( lh.get_intersection_axes(), [&]( const vec< N >& axis_direction ) {
                       return axis_separates< true >( lh, rh, axis_direction );
               } );
}

/* Checks collision between provided objects. The Separation Axis Test is used. This simply calls
 * the collidesSingleAxes(A,B) on the axes of 'RH' object and than on the axes of 'RH' object.
 *
 * O( (x+z)*(k+n) ), where x,z are the number of axes for the LH and RH objects and k,n are the
 * number of points for LH and RH.
 */
template < std::size_t N, typename LH, typename RH >
[[gnu::hot]] constexpr bool collides( const LH& lh, const RH& rh ) noexcept
{
        static_assert( std::is_same_v< typename LH::tag, typename RH::tag > );
        static_assert( std::is_base_of_v< i_collision_item< typename LH::point_iterator,
                                                            typename LH::separation_axis_iterator >,
                                          LH > );
        static_assert( std::is_base_of_v< i_collision_item< typename RH::point_iterator,
                                                            typename RH::separation_axis_iterator >,
                                          RH > );
        return collides_single_axes< LH, RH, N >( lh, rh ) &&
               collides_single_axes< RH, LH, N >( rh, lh );
}

/* Fast check of collision between provided objects. This uses min-spheres for each object. The min
 * sphere is the smallest sphere containing adequate object. The collision check between two spheres
 * is easy - simply compare the distance between their centers with their radiuses.
 *
 * O(1)
 */
template < std::size_t N, typename LH, typename RH >
[[gnu::hot]] constexpr bool fast_collides( const LH& lh, const RH& rh ) noexcept
{
        static_assert(
            std::is_base_of_v< i_fast_collision_item< typename LH::full_type, N >, LH > );
        static_assert(
            std::is_base_of_v< i_fast_collision_item< typename RH::full_type, N >, RH > );
        static_assert( std::is_same_v< typename LH::tag, typename RH::tag > );

        em::distance dist{
            distance_of( lh.get_min_sphere_position(), rh.get_min_sphere_position() ) };

        return dist < ( lh.get_min_sphere_radius() + rh.get_min_sphere_radius() );
}

template < std::size_t    N,
           collision_mode Mode = collision_mode::FAST,
           typename LH,
           typename RH,
           typename UnaryFunction >
[[gnu::hot]] constexpr bool trees_nodes_collides( const LH&     lh,
                                                  const RH&     rh,
                                                  UnaryFunction f ) noexcept( noexcept( f( lh,
                                                                                           rh ) ) )
{
        // check if lh and rh collides with each other
        if constexpr ( Mode == collision_mode::FULL ) {
                if ( !collides< N >( lh, rh ) ) {
                        return false;
                }
        } else if constexpr ( Mode == collision_mode::FAST ) {
                if ( !fast_collides< N >( lh, rh ) ) {
                        return false;
                }
        }

        // check leafness
        bool step_down_right = false;

        if ( lh.is_leaf() && rh.is_leaf() ) {
                if constexpr ( Mode == collision_mode::FAST ) {
                        return collides< N >( lh.get_full(), rh.get_full() ) && f( lh, rh );
                } else {
                        return f( lh, rh );
                }
        } else if ( lh.is_leaf() ) {
                step_down_right = true;
        } else if ( rh.is_leaf() ) {
                step_down_right = false;
        } else {
                if constexpr ( Mode == collision_mode::FAST ) {
                        step_down_right = rh.get_min_sphere_radius() > lh.get_min_sphere_radius();
                } else {
                        step_down_right = rh > lh;
                }
        }

        // step down either to the right or to the left
        if ( step_down_right ) {
                return em::any_of( rh.get_children(), [&]( const auto& child ) {  //
                        return trees_nodes_collides< N, Mode >( lh, child, f );
                } );
        }
        return em::any_of( lh.get_children(), [&]( const auto& child ) {  //
                return trees_nodes_collides< N, Mode >( child, rh, f );
        } );
}

template < std::size_t    N,
           collision_mode Mode,
           typename LH,
           typename RH,
           typename CollisionFunction,
           typename CollectFunction >
[[gnu::hot]] constexpr void find_collision_nodes( const LH&         lh,
                                                  const RH&         rh,
                                                  CollisionFunction collision_f,
                                                  CollectFunction   collect_f )
{
        trees_nodes_collides< N, Mode >( lh, rh, [&]( const auto& lh_node, const auto& rh_node ) {
                if ( !collision_f( lh_node, rh_node ) ) {
                        return false;
                }

                collect_f( lh_node, rh_node );

                return false;
        } );
}

}  // namespace schpin
