#include "base.h"
#include "schpin_lib/algorithm.h"
#include "schpin_lib/util.h"
#include "util.h"

#include <array>
#include <memory>
#include <numeric>
#include <utility>

#pragma once

namespace schpin
{

template < typename Content >
class oc_node
{
      public:
        using container      = std::array< oc_node, 8 >;
        using iterator       = oc_node*;
        using const_iterator = const oc_node*;

        explicit oc_node( std::unique_ptr< container >&& nodes, Content content )
          : children_ptr_( std::move( nodes ) )
          , content_( std::move( content ) )
        {
        }

        explicit oc_node( Content content = Content{} )
          : children_ptr_( nullptr )
          , content_( std::move( content ) )
        {
        }

        oc_node copy() const
        {
                std::unique_ptr< container > ptr;
                if ( children_ptr_ ) {
                        ptr = std::make_unique< container >(
                            em::map_f_to_a< 8 >( em::range( 8u ), [&]( std::size_t i ) {  //
                                    return ( *children_ptr_ )[i].copy();
                            } ) );
                }
                return oc_node( std::move( ptr ), content_ );
        }

        template < typename Proxy, typename... Args >
        Proxy overlay_children( std::size_t index, Args... args )
        {
                return Proxy( children_ptr_->at( index ), args... );
        }

        constexpr std::size_t size() const
        {
                if ( children_ptr_ ) {
                        return children_ptr_->size();
                }
                return 0;
        }

        constexpr bool has_children() const
        {
                return bool( children_ptr_ );
        }

        constexpr em::view< const_iterator > get_children() const
        {
                if ( children_ptr_ ) {
                        return em::view< const_iterator >( children_ptr_->begin(),
                                                           children_ptr_->end() );
                }
                return em::view< const_iterator >( nullptr, nullptr );
        }

        constexpr bool is_leaf() const
        {
                return !bool( children_ptr_ );
        }

        constexpr Content& operator*()
        {
                return content_;
        }

        constexpr Content* operator->()
        {
                return &content_;
        }
        constexpr const Content& operator*() const
        {
                return content_;
        }
        constexpr const Content* operator->() const
        {
                return &content_;
        }

        void init_children()
        {
                container children{ oc_node< Content >( content_ ),
                                    oc_node< Content >( content_ ),
                                    oc_node< Content >( content_ ),
                                    oc_node< Content >( content_ ),
                                    oc_node< Content >( content_ ),
                                    oc_node< Content >( content_ ),
                                    oc_node< Content >( content_ ),
                                    oc_node< Content >( content_ ) };

                children_ptr_ = std::make_unique< container >( std::move( children ) );
        }

        void delete_children()
        {
                children_ptr_.reset( nullptr );
        }

      private:
        std::unique_ptr< container > children_ptr_{};
        Content                      content_;
};

template < typename Content >
constexpr bool operator!=( const oc_node< Content >& lh, const oc_node< Content >& rh );

template < typename Content >
constexpr bool operator==( const oc_node< Content >& lh, const oc_node< Content >& rh )
{
        if ( *lh != *rh ) {
                return false;
        }

        return em::all_of( zip( lh.get_children(), rh.get_children() ),
                           em::uncurry | std::equal_to< oc_node< Content > >() );
}

template < typename Content >
constexpr bool operator!=( const oc_node< Content >& lh, const oc_node< Content >& rh )
{
        return !( lh == rh );
}

template < typename Content >
constexpr uint64_t get_node_count( const oc_node< Content >& node )
{
        return get_node_count( node, []( const auto& ) { return true; } );
}
template < typename Content, typename Predicate >
constexpr uint64_t get_node_count( const oc_node< Content >& node, Predicate&& pred )
{
        return ( pred( node ) ? 1u : 0u ) +
               em::sum( node.get_children(), [&]( const oc_node< Content >& node ) {  //
                       return get_node_count( node, pred );
               } );
}

template < typename Content >
constexpr uint64_t get_tree_height( const oc_node< Content >& node )
{
        return em::max_elem( node.get_children(),
                             [&]( const auto& n ) { return get_tree_height( n ); } ) +
               1;
}

template < typename Content >
constexpr uint64_t get_memory_usage( const oc_node< Content >& node )
{
        return node.get_content().get_memory_usage( node ) +
               em::sum( node.get_children(), get_memory_usage );
}

}  // namespace schpin

template < typename Content >
struct std::hash< schpin::oc_node< Content > >
{
        std::size_t operator()( const schpin::oc_node< Content >& node )
        {
                std::string as_str =
                    std::to_string( std::hash< Content >()( *node ) ) +
                    em::sum( node.get_children(), [&]( const schpin::oc_node< Content >& child ) {
                            return std::to_string(
                                std::hash< schpin::oc_node< Content > >()( child ) );
                    } );
                return std::hash< std::string >()( as_str );
        }
};
