#include "base/physics.h"
#include "octree/base.h"
#include "octree/interfaces.h"
#include "octree/node.h"
#include "octree/proxy_node.h"
#include "octree/tree.h"

#include <nlohmann/json.hpp>

#pragma once

namespace schpin
{

constexpr occupancy oc_default_occupancy( 0.5f );
constexpr occupancy oc_prunning_diff_occupancy( 0.000001f );
constexpr occupancy oc_min_occupancy( 0.1192f );
constexpr occupancy oc_max_occupancy( 0.971f );

enum class occupancy_token
{
        FULL
};

class occupancy_content : public i_oc_content< occupancy_content, occupancy_token >
{
        occupancy oc_;

      public:
        occupancy_content() = default;

        template < typename Item >
        explicit occupancy_content( Item def )
          : oc_( def )
        {
        }

        constexpr const occupancy& get_occupancy() const
        {
                return oc_;
        }

        std::size_t get_memory_usage( const oc_node< occupancy_content >& node ) const final
        {
                return sizeof( node );
        }

        bool prunnable( const oc_node< occupancy_content >& ) const final
        {
                // shouldn't we check for children count here? or something? like all children are
                // leafs?
                return abs( oc_max_occupancy - oc_ ) < oc_prunning_diff_occupancy ||
                       abs( oc_min_occupancy - oc_ ) < oc_prunning_diff_occupancy;
        }
        void child_based_update( const oc_node< occupancy_content >& node ) final
        {
                oc_ =
                    em::sum( node.get_children(), []( const oc_node< occupancy_content >& child ) {
                            return child->get_occupancy();
                    } );
                oc_ /= float( node.size() );
        }

        void set_prunned( const oc_node< occupancy_content >& ) final
        {
                if ( oc_ > ( oc_max_occupancy + oc_min_occupancy ) / 2.f ) {
                        oc_ = oc_max_occupancy;
                } else {
                        oc_ = oc_min_occupancy;
                }
        }

        void update( const occupancy_token& token ) final
        {
                switch ( token ) {
                        case occupancy_token::FULL:
                                oc_ = oc_max_occupancy;
                                return;
                }
        }
};

struct occupancy_treshold_predicate
{
        constexpr bool operator()( const oc_node< occupancy_content >& node )
        {
                return node->get_occupancy() > occupancy( 0.5f );
        }

        template < typename Tag >
        constexpr bool operator()( const oc_proxy_node< occupancy_content, Tag >& node )
        {
                return this->operator()( *node );
        }
};

constexpr bool operator==( const occupancy_content& lh, const occupancy_content& rh )
{
        return lh.get_occupancy() == rh.get_occupancy();
}
constexpr bool operator!=( const occupancy_content& lh, const occupancy_content& rh )
{
        return !( lh == rh );
}

inline void from_json( const nlohmann::json& j, occupancy_content& c )
{
        c = occupancy_content{ j.get< float >() };
}

inline void to_json( nlohmann::json& j, const occupancy_content& c )
{
        j = *( c.get_occupancy() );
}

struct oc_tree_binary_coder
{
        using node = oc_node< occupancy_content >;

        std::vector< uint8_t > data;
        std::size_t            i = 0;

        void binary_encode( const node& inpt )
        {
                data.push_back( inpt.has_children() ? 1 : 0 );
                float val = *inpt->get_occupancy();
                std::copy_n( reinterpret_cast< uint8_t* >( &val ),
                             sizeof( val ),
                             std::back_inserter( data ) );

                for ( const auto& child : inpt.get_children() ) {
                        binary_encode( child );
                }
        }

        node binary_decode()
        {
                std::size_t j = i;
                i += 1;
                float val = 0;
                std::copy_n( &data[j + 1], sizeof( val ), reinterpret_cast< uint8_t* >( &val ) );
                std::unique_ptr< std::array< node, 8 > > ptr;
                if ( data[j] != 0u ) {
                        std::array< node, 8 > nodes{ binary_decode(),
                                                     binary_decode(),
                                                     binary_decode(),
                                                     binary_decode(),
                                                     binary_decode(),
                                                     binary_decode(),
                                                     binary_decode(),
                                                     binary_decode() };
                        ptr = std::make_unique< std::array< node, 8 > >( std::move( nodes ) );
                }
                return oc_node{ std::move( ptr ), occupancy_content{ occupancy{ val } } };
        };
};

}  // namespace schpin

template <>
struct std::hash< schpin::occupancy_content >
{
        std::size_t operator()( const schpin::occupancy_content& content )
        {
                return std::hash< schpin::occupancy >()( content.get_occupancy() );
        }
};
