#include "schpin_lib/geom/pose.h"

#include <interactive_markers/interactive_marker_server.h>
#include <tf2_ros/transform_broadcaster.h>
#include <utility>

#pragma once

namespace schpin
{

visualization_msgs::InteractiveMarker make_pose_marker( const std::string& name, std::string frame )
{
        visualization_msgs::InteractiveMarker int_marker;
        int_marker.header.frame_id = std::move( frame );
        int_marker.name            = name + "_marker";
        int_marker.description     = name + " marker";
        int_marker.scale           = 0.1f;

        visualization_msgs::Marker box_marker;
        box_marker.frame_locked = 1u;
        box_marker.type         = visualization_msgs::Marker::SPHERE;
        box_marker.scale.x      = 0.05;
        box_marker.scale.y      = 0.05;
        box_marker.scale.z      = 0.05;
        box_marker.color.r      = 0.5;
        box_marker.color.g      = 0.5;
        box_marker.color.b      = 0.5;
        box_marker.color.a      = 0.5;

        visualization_msgs::InteractiveMarkerControl box_control;
        box_control.always_visible = 1u;
        box_control.markers.push_back( box_marker );

        int_marker.controls.push_back( box_control );

        visualization_msgs::InteractiveMarkerControl control;
        control.always_visible = 1u;

        control.orientation.w    = 1;
        control.orientation.x    = 1;
        control.orientation.y    = 0;
        control.orientation.z    = 0;
        control.name             = name + "_move_x";
        control.interaction_mode = visualization_msgs::InteractiveMarkerControl::MOVE_AXIS;
        int_marker.controls.push_back( control );
        control.name             = name + "_rotate_x";
        control.interaction_mode = visualization_msgs::InteractiveMarkerControl::ROTATE_AXIS;
        int_marker.controls.push_back( control );

        control.orientation.w    = 1;
        control.orientation.x    = 0;
        control.orientation.y    = 1;
        control.orientation.z    = 0;
        control.name             = name + "_move_y";
        control.interaction_mode = visualization_msgs::InteractiveMarkerControl::MOVE_AXIS;
        int_marker.controls.push_back( control );
        control.name             = name + "_rotate_y";
        control.interaction_mode = visualization_msgs::InteractiveMarkerControl::ROTATE_AXIS;
        int_marker.controls.push_back( control );

        control.orientation.w    = 1;
        control.orientation.x    = 0;
        control.orientation.y    = 0;
        control.orientation.z    = 1;
        control.name             = name + "_move_z";
        control.interaction_mode = visualization_msgs::InteractiveMarkerControl::MOVE_AXIS;
        int_marker.controls.push_back( control );
        control.name             = name + "_rotate_z";
        control.interaction_mode = visualization_msgs::InteractiveMarkerControl::ROTATE_AXIS;
        int_marker.controls.push_back( control );

        return int_marker;
}

class interactive_pose
{
        interactive_markers::InteractiveMarkerServer server_;
        std::string                                  name_;
        pose< world_frame >                          pose_;
        std::string                                  frame_;
        tf2_ros::TransformBroadcaster                br_;

      public:
        interactive_pose( interactive_pose&& other ) noexcept
          : server_( other.name_ )
          , name_( std::move( other.name_ ) )
          , pose_( other.pose_ )
          , frame_( std::move( other.frame_ ) )
          , br_( other.br_ )
        {
        }

        interactive_pose( const std::string& name, std::string frame )
          : server_( name )
          , name_( name )
          , frame_( std::move( frame ) )
        {
                show();
                set_tf();
        }

        const pose< world_frame >& get_pose() const
        {
                return pose_;
        }

        void set_pose( const pose< world_frame >& p )
        {
                pose_ = p;
                set_tf();
        }

      private:
        void show()
        {
                auto marker = make_pose_marker( name_, "world" );
                server_.insert(
                    marker,
                    [this](
                        const visualization_msgs::InteractiveMarkerFeedbackConstPtr& feedback ) {
                            this->on_feedback( feedback );
                    } );
                server_.applyChanges();
        }

        void on_feedback( const visualization_msgs::InteractiveMarkerFeedbackConstPtr& feedback )
        {

                pose_ = msg_to_pose< world_frame >( feedback->pose );
                set_tf();
        }
        void set_tf()
        {
                geometry_msgs::TransformStamped transform_stamped;

                transform_stamped.header.stamp          = ros::Time::now();
                transform_stamped.header.frame_id       = "world";
                transform_stamped.child_frame_id        = frame_;
                transform_stamped.transform.translation = vector_to_msg( pose_.position );
                transform_stamped.transform.rotation    = quaternion_to_msg( pose_.orientation );
                br_.sendTransform( transform_stamped );
        }
};

}  // namespace schpin
